from asphalt.core.context import Context
import pytest

from asphalt.mailer.api import Mailer
from asphalt.mailer.component import MailerComponent


@pytest.mark.parametrize('backend', ['smtp', 'sendmail'])
@pytest.mark.asyncio
async def test_component_single(caplog, backend):
    component = MailerComponent(backend=backend)
    context = Context()
    await component.start(context)

    assert isinstance(context.mailer, Mailer)
    records = [record for record in caplog.records if record.name == 'asphalt.mailer.component']
    records.sort(key=lambda r: r.message)
    assert len(records) == 1
    assert records[0].message.startswith("Configured mailer (default / ctx.mailer)")


@pytest.mark.asyncio
async def test_component_multiple(caplog):
    component = MailerComponent(mailers={
        'smtp': {'backend': 'smtp', 'context_attr': 'mailer1'},
        'sendmail': {'backend': 'sendmail', 'context_attr': 'mailer2'}
    })
    context = Context()
    await component.start(context)

    assert isinstance(context.mailer1, Mailer)
    assert isinstance(context.mailer2, Mailer)
    records = [record for record in caplog.records if record.name == 'asphalt.mailer.component']
    records.sort(key=lambda r: r.message)
    assert len(records) == 2
    assert records[0].message.startswith("Configured mailer (sendmail / ctx.mailer2)")
    assert records[1].message.startswith("Configured mailer (smtp / ctx.mailer1)")



#!/usr/bin/env python3
import pkg_resources


extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx_autodoc_typehints',
    'sphinxcontrib.asyncio'
]

templates_path = ['_templates']
source_suffix = '.rst'
master_doc = 'index'
project = 'asphalt-mailer'
author = 'Alex Grönholm'
copyright = '2015, ' + author

v = pkg_resources.get_distribution(project).parsed_version
version = v.base_version
release = v.public

language = None

exclude_patterns = ['_build']
pygments_style = 'sphinx'
highlight_language = 'python3'
todo_include_todos = False

html_theme = 'classic'
html_static_path = ['_static']
htmlhelp_basename = project.replace('-', '') + 'doc'

intersphinx_mapping = {'python': ('http://docs.python.org/3/', None),
                       'asphalt': ('http://asphalt.readthedocs.org/en/latest/', None)}



from email.message import EmailMessage
from typing import Iterable, Union, Dict, Any

from typeguard import check_argument_types

from asphalt.mailer.api import Mailer


class MockMailer(Mailer):
    """
    A mailer that does not send any messages but instead stores them in a member variable.

    :param message_defaults: default values for omitted keyword arguments of
        :meth:`~asphalt.mailer.api.Mailer.create_message`

    :ivar messages: list of messages that would normally have been sent
    """

    __slots__ = 'messages'

    def __init__(self, *, message_defaults: Dict[str, Any]=None):
        assert check_argument_types()
        super().__init__(message_defaults or {})
        self.messages = []

    async def deliver(self, messages: Union[EmailMessage, Iterable[EmailMessage]]):
        assert check_argument_types()
        if isinstance(messages, EmailMessage):
            messages = [messages]

        self.messages.extend(messages)

    def __repr__(self):
        return '{0.__class__.__name__}()'.format(self)



from email.message import EmailMessage

import pytest


@pytest.fixture
def sample_message():
    msg = EmailMessage()
    msg['From'] = 'foo@bar.baz'
    msg['To'] = 'Test Recipient <test@domain.country>, test2@domain.country'
    msg['Cc'] = 'Test CC <testcc@domain.country>, testcc2@domain.country'
    msg['Bcc'] = 'Test BCC <testbcc@domain.country>, testbcc2@domain.country'
    msg.set_content('Test content')
    return msg


@pytest.fixture
def recipients():
    return ('test@domain.country', 'test2@domain.country',
            'testcc@domain.country', 'testcc2@domain.country',
            'testbcc@domain.country', 'testbcc2@domain.country')



from email.headerregistry import Address
from email.message import EmailMessage
from typing import Union, Iterable

import pytest

from asphalt.mailer.api import Mailer


class DummyMailer(Mailer):
    def __init__(self, **message_defaults):
        super().__init__(message_defaults)
        self.messages = []

    async def deliver(self, messages: Union[EmailMessage, Iterable[EmailMessage]]):
        messages = [messages] if isinstance(messages, EmailMessage) else messages
        self.messages.extend(messages)


@pytest.fixture
def kwargs():
    return {
        'charset': 'iso-8859-1',
        'plain_body': 'Hello åäö',
        'html_body': '<html><body>Hello åäö</body></html>',
        'to': [Address('Test Recipient', 'test', 'domain.country'), 'test2@domain.country'],
        'cc': [Address('Test CC', 'testcc', 'domain.country'), 'testcc2@domain.country'],
        'bcc': [Address('Test BCC', 'testbcc', 'domain.country'), 'testbcc2@domain.country']
    }


@pytest.fixture
def mailer():
    return DummyMailer()


@pytest.mark.parametrize('plain_body, html_body', [
    (True, True), (True, False), (False, True)
], ids=['both', 'plain', 'html'])
def test_create_message(mailer, kwargs, plain_body, html_body):
    if not plain_body:
        del kwargs['plain_body']
    if not html_body:
        del kwargs['html_body']
    msg = mailer.create_message(subject='test subject', sender='foo@bar.baz', **kwargs)

    assert msg['From'] == 'foo@bar.baz'
    assert msg['Subject'] == 'test subject'
    assert msg['To'] == 'Test Recipient <test@domain.country>, test2@domain.country'
    assert msg['Cc'] == 'Test CC <testcc@domain.country>, testcc2@domain.country'
    assert msg['Bcc'] == 'Test BCC <testbcc@domain.country>, testbcc2@domain.country'

    if plain_body and html_body:
        assert msg['Content-Type'] == 'multipart/alternative'
        plain_part, html_part = msg.iter_parts()
    elif plain_body:
        plain_part, html_part = msg, None
    else:
        plain_part, html_part = None, msg

    if plain_part:
        assert plain_part['Content-Type'] == 'text/plain; charset="iso-8859-1"'
        assert plain_part.get_content() == 'Hello åäö\n'
    if html_part:
        assert html_part['Content-Type'] == 'text/html; charset="iso-8859-1"'
        assert html_part.get_content() == '<html><body>Hello åäö</body></html>\n'


def test_message_defaults():
    """
    Test that message defaults are applied when the corresponding arguments have been omitted.

    """
    mailer = DummyMailer(subject='default_subject', sender='default_sender@example.org',
                         to='default_recipient@example.org', cc='default_cc@example.org',
                         bcc='default_bcc@example.org', charset='utf-16')
    msg = mailer.create_message(plain_body='Hello åäö')
    assert msg['Subject'] == 'default_subject'
    assert msg['From'] == 'default_sender@example.org'
    assert msg['To'] == 'default_recipient@example.org'
    assert msg['Cc'] == 'default_cc@example.org'
    assert msg['Bcc'] == 'default_bcc@example.org'
    assert msg.get_charsets() == ['utf-16']


def test_add_attachment(mailer):
    msg = mailer.create_message(subject='foo')
    mailer.add_attachment(msg, b'binary content', filename='test')
    attachments = list(msg.iter_attachments())
    assert len(attachments) == 1
    assert attachments[0]['Content-Type'] == 'application/octet-stream'
    assert attachments[0]['Content-Disposition'] == 'attachment; filename="test"'


@pytest.mark.asyncio
async def test_add_file_attachment(mailer):
    msg = mailer.create_message(subject='foo')
    await mailer.add_file_attachment(msg, __file__)
    attachments = list(msg.iter_attachments())
    assert len(attachments) == 1
    assert attachments[0]['Content-Type'] == 'text/x-python'
    assert attachments[0]['Content-Disposition'] == 'attachment; filename="test_api.py"'


def test_add_attachment_bad_mime_type(mailer):
    msg = mailer.create_message(subject='foo')
    exc = pytest.raises(ValueError, mailer.add_attachment, msg, b'abc', 'file.dat', '/badtype')
    assert str(exc.value) == 'mimetype must be a string in the "maintype/subtype" format'


@pytest.mark.asyncio
async def test_create_and_deliver(mailer, kwargs):
    await mailer.create_and_deliver(subject='test subject', sender='foo@bar.baz', **kwargs)
    assert len(mailer.messages) == 1
    assert isinstance(mailer.messages[0], EmailMessage)
    assert mailer.messages[0]['From'] == 'foo@bar.baz'



import subprocess
import sys
from asyncio import create_subprocess_exec
from email.message import EmailMessage
from pathlib import Path
from typing import Iterable, Union, Dict, Any

from typeguard import check_argument_types

from asphalt.mailer.api import Mailer, DeliveryError
from asphalt.mailer.util import get_recipients

__all__ = ['SendmailMailer']


class SendmailMailer(Mailer):
    """
    A mailer that sends mail by running the ``sendmail`` executable in a subprocess.

    :param path: path to the sendmail executable
    :param message_defaults: default values for omitted keyword arguments of
        :meth:`~asphalt.mailer.api.Mailer.create_message`
    """

    __slots__ = 'path'

    def __init__(self, *, path: Union[str, Path] = '/usr/sbin/sendmail',
                 message_defaults: Dict[str, Any] = None):
        assert check_argument_types()
        super().__init__(message_defaults or {})
        self.path = str(path)

    async def deliver(self, messages: Union[EmailMessage, Iterable[EmailMessage]]):
        assert check_argument_types()
        if isinstance(messages, EmailMessage):
            messages = [messages]

        for message in messages:
            args = [self.path, '-i', '-B', '8BITMIME'] + get_recipients(message)
            try:
                process = await create_subprocess_exec(*args, stdin=subprocess.PIPE,
                                                       stderr=subprocess.PIPE)
            except Exception as e:
                raise DeliveryError(str(e), message) from e

            del message['Bcc']
            stdout, stderr = await process.communicate(message.as_bytes())
            if process.returncode:
                error = stderr.decode(sys.stderr.encoding).rstrip()
                raise DeliveryError(error, message)

    def __repr__(self):
        return '{0.__class__.__name__}({0.path!r})'.format(self)



from email.message import EmailMessage
from typing import List

from typeguard import check_argument_types


def get_recipients(message: EmailMessage) -> List[str]:
    """
    Return a list of email addresses of all the intended recipients of the given message.

    This function is meant to be used by :class:`~asphalt.mailer.api.Mailer` implementations.

    :param message: the source email message

    """
    assert check_argument_types()
    recipients = []
    for header in (message['To'], message['Cc'], message['Bcc']):
        if header:
            for addr in header.addresses:
                recipients.append(addr.addr_spec)

    return recipients



import platform
import re
from asyncio.streams import StreamWriter, StreamReader, open_connection
from asyncio.tasks import wait_for
from base64 import b64encode
from email.message import EmailMessage
from email.policy import SMTP
from ssl import SSLContext
from typing import Iterable, Union, Dict, Any

from asyncio_extras.asyncyield import yield_async
from asyncio_extras.contextmanager import async_contextmanager
from typeguard import check_argument_types

from asphalt.core.context import Context
from asphalt.core.util import resolve_reference
from asphalt.mailer.api import Mailer, DeliveryError
from asphalt.mailer.util import get_recipients

__all__ = 'SMTPError', 'SMTPMailer'


class SMTPError(DeliveryError):
    """
    Raised when the SMTP server returns a response with a code of 400 or greater.

    :param code: the numeric response code
    :param error: the error message from the server
    """

    def __init__(self, code: int, error: str):
        super().__init__('server returned error: {} {}'.format(code, error))
        self.code = code
        self.error = error


class SMTPConnection:
    __slots__ = 'reader', 'writer', 'timeout'

    response_re = re.compile(r'(\d{3})(-| )(.*)')

    def __init__(self, reader: StreamReader, writer: StreamWriter, timeout: int):
        self.reader = reader
        self.writer = writer
        self.timeout = timeout

    @classmethod
    @async_contextmanager
    async def connect(cls, host, port, ssl, timeout):
        reader, writer = await wait_for(open_connection(host, port, ssl=ssl), timeout)
        connection = cls(reader, writer, timeout)
        await yield_async(connection)
        writer.close()

    async def command(self, command: str, *args: str, expect_response: bool=True):
        # Send the command
        data = ' '.join((command,) + args).encode('ascii') + b'\r\n'
        self.write(data)
        if expect_response:
            return await self.get_response()

    async def get_response(self):
        lines = []
        while True:
            read_task = self.reader.readline()
            line = await wait_for(read_task, self.timeout)
            if not line.endswith(b'\r\n'):
                raise DeliveryError('server closed connection')

            line = line.decode(errors='replace').rstrip()
            match = self.response_re.match(line)
            if not match:
                raise DeliveryError('SMTP protocol error: received unexpected response: {}'.
                                    format(line))

            code, separator, message = match.groups()
            lines.append(message)
            if separator == ' ':
                code = int(code)
                response = '\n'.join(lines)
                if code >= 400:
                    raise SMTPError(code, response)
                return response

    def write(self, data: bytes):
        self.writer.write(data)


class SMTPMailer(Mailer):
    """
    A mailer that uses the ESMTP protocol (:rfc:`2821`) to send mails.

    Supports only unencrypted or implicit TLS connections. STARTTLS is not supported yet.

    :param host: the host name or IP address of the SMTP server to connect to
    :param port: the port number to connect to (omit to autodetect based on the ``ssl`` parameter)
    :param ssl: one of the following:

        * ``False`` to disable SSL
        * ``True`` to enable SSL using the default context
        * an :class:`ssl.SSLContext` instance
        * a module:varname reference to an :class:`ssl.SSLContext` instance
        * name of an :class:`ssl.SSLContext` resource
    :param username: user name to authenticate as
    :param password: password to authenticate with
    :param timeout: timeout (in seconds) for connect and read operations
    :param message_defaults: default values for omitted keyword arguments of
        :meth:`~asphalt.mailer.api.Mailer.create_message`
    """

    __slots__ = 'host', 'port', 'ssl', 'username', 'password', 'timeout'

    def __init__(self, *, host: str = '127.0.0.1', port: int = None,
                 ssl: Union[bool, str, SSLContext] = False, username: str = None,
                 password: str = None, timeout: Union[int, float] = 10,
                 message_defaults: Dict[str, Any] = None):
        assert check_argument_types()
        super().__init__(message_defaults)
        self.host = host
        self.ssl = resolve_reference(ssl)
        self.port = port or ('465' if self.ssl else '25')
        self.username = username
        self.password = password
        self.timeout = timeout

    async def start(self, ctx: Context):
        if isinstance(self.ssl, str):
            self.ssl = await ctx.request_resource(SSLContext, self.ssl)

    async def deliver(self, messages: Union[EmailMessage, Iterable[EmailMessage]]):
        assert check_argument_types()
        if isinstance(messages, EmailMessage):
            messages = [messages]

        async with SMTPConnection.connect(self.host, self.port, self.ssl,
                                          self.timeout) as connection:
            await connection.get_response()
            response = await connection.command('EHLO', platform.node())
            features = response.split('\n')[1:]
            policy = SMTP.clone(cte_type='8bit' if '8BITMIME' in features else '7bit')
            pipelining = 'PIPELINING' in features

            try:
                # Authenticate if necessary
                if self.username and self.password:
                    auth_mechanisms = []
                    for feature in features:
                        if feature.startswith('AUTH '):
                            auth_mechanisms = feature[5:].split()
                            break

                    if 'PLAIN' in auth_mechanisms:
                        token = b64encode('\0{}\0{}'.format(self.username, self.password).encode())
                        await connection.command('AUTH PLAIN', token.decode())
                    elif 'LOGIN' in auth_mechanisms:
                        await connection.command('AUTH LOGIN', self.username, self.password)
                    else:
                        raise DeliveryError('server does not support any of our authentication '
                                            'methods')

                # Send each message in a separate mail transaction
                for message in messages:
                    try:
                        recipients = get_recipients(message)
                        await connection.command(
                            'MAIL FROM: <{}>'.format(message['From'].addresses[0].addr_spec),
                            expect_response=not pipelining)
                        for recipient in recipients:
                            await connection.command('RCPT TO: <{}>'.format(recipient),
                                                     expect_response=not pipelining)

                        await connection.command('DATA', expect_response=not pipelining)

                        # If pipelining was in use, the responses to the MAIL, RCPT and DATA
                        # commands will be sent by the server now
                        if pipelining:
                            for _ in range(2 + len(recipients)):
                                await connection.get_response()

                        envelope = message.as_bytes(policy=policy)
                        connection.write(envelope + b'\r\n')
                        await connection.command('.')
                    except DeliveryError as e:
                        e.message = message
                        raise
            except DeliveryError:
                await connection.command('QUIT', expect_response=False)
                raise
            else:
                await connection.command('QUIT', expect_response=False)

    def __repr__(self):
        return '{self.__class__.__name__}(host={self.host!r}, port={self.port!r})'.\
            format(self=self)



import logging
from typing import Dict, Any

from typeguard import check_argument_types

from asphalt.core.component import Component
from asphalt.core.context import Context
from asphalt.core.util import PluginContainer, merge_config
from asphalt.mailer.api import Mailer

mailer_backends = PluginContainer('asphalt.mailer.mailers', Mailer)

logger = logging.getLogger(__name__)


class MailerComponent(Component):
    """
    Publishes one or more :class:`~asphalt.mailer.api.Mailer` resources.

    If more than one mailer is to be configured, provide a ``mailers`` argument as a dictionary
    where the key is the resource name and the value is a dictionary of keyword arguments to
    :meth:`create_mailer`. Otherwise, directly pass those keyword arguments to the component
    constructor itself.

    If ``mailers`` is defined, any extra keyword arguments are used as default values for
    :meth:`create_mailer` for all mailers (:func:`~asphalt.core.util.merge_config` is used to
    merge the per-mailer arguments with the defaults). Otherwise, a single mailer is created based
    on the provided default arguments, with ``context_attr`` defaulting to ``mailer``.

    :param mailers: a dictionary of resource name ⭢ :meth:`create_mailer` keyword arguments
    :param default_mailer_args: default values for :meth:`create_mailer`
    """

    def __init__(self, mailers: Dict[str, Dict[str, Any]] = None, **default_mailer_args):
        assert check_argument_types()
        mailers = mailers or {}
        if default_mailer_args:
            default_mailer_args.setdefault('context_attr', 'mailer')
            mailers['default'] = default_mailer_args

        self.mailers = []
        for resource_name, config in mailers.items():
            config = merge_config(default_mailer_args, config)
            config.setdefault('context_attr', resource_name)
            context_attr, mailer = self.create_mailer(**config)
            self.mailers.append((resource_name, context_attr, mailer))

    @classmethod
    def create_mailer(cls, context_attr: str, backend: str, **backend_kwargs):
        """
        Configure a Mailer backend with the given parameters.

        :param context_attr: the mailer's attribute name on the context (defaults to the resource
            name)
        :param backend: specifies the type of mailer (declared as entry points in the
            ``asphalt.mailer.mailers`` namespace)
        :param backend_kwargs: keyword arguments passed to the constructor of the backend class

        """
        assert check_argument_types()
        mailer = mailer_backends.create_object(backend, **backend_kwargs)
        return context_attr, mailer

    async def start(self, ctx: Context):
        for resource_name, context_attr, mailer in self.mailers:
            await mailer.start(ctx)
            ctx.publish_resource(mailer, resource_name, context_attr, types=[Mailer])
            logger.info('Configured mailer (%s / ctx.%s)', resource_name, context_attr)



from email.message import EmailMessage

from asphalt.mailer.util import get_recipients


def test_get_recipients():
    msg = EmailMessage()
    msg['To'] = 'Foo Example <foo@example.org>'
    msg['Cc'] = 'Bar Bar <bar@bar.bar>'
    msg['Bcc'] = 'Invisible Recipient <invisible@reci.pient>'

    assert get_recipients(msg) == ['foo@example.org', 'bar@bar.bar', 'invisible@reci.pient']



from abc import abstractmethod, ABCMeta
from collections import Awaitable
from email.headerregistry import Address
from email.message import EmailMessage
from mimetypes import guess_type
from pathlib import Path
from typing import Iterable, Union, Dict, Any

from asyncio_extras.threads import call_in_executor
from typeguard import check_argument_types

from asphalt.core.context import Context

__all__ = ('DeliveryError', 'Mailer')

AddressListType = Union[str, Address, Iterable[Union[str, Address]]]


class DeliveryError(Exception):
    """
    Raised when there's an error with mail delivery.

    :ivar error: the error message
    :ivar message: the email message related to the failure, if any
    """

    def __init__(self, error: str, message: EmailMessage = None):
        super().__init__(error, message)

    def __str__(self):
        return 'error sending mail message: {}'.format(self.args[0])


class Mailer(metaclass=ABCMeta):
    """
    This is the abstract base class for all mailers.

    :param message_defaults: default values for omitted keyword arguments of :meth:`create_message`
    """

    __slots__ = 'message_defaults'

    def __init__(self, message_defaults: Dict[str, Any] = None):
        self.message_defaults = message_defaults or {}
        self.message_defaults.setdefault('charset', 'utf-8')

    async def start(self, ctx: Context):
        """
        Perform any necessary setup procedures.

        This method is called by the component on initialization.

        :param ctx: the mailer component's context
        """

    def create_message(self, *, subject: str = None, sender: Union[str, Address] = None,
                       to: AddressListType = None, cc: AddressListType = None,
                       bcc: AddressListType = None, charset: str = None, plain_body: str = None,
                       html_body: str = None) -> EmailMessage:
        """
        Create an :class:`~email.message.EmailMessage` using to be sent later using
        :meth:`deliver`.

        :param subject: subject line for the message
        :param sender: sender address displayed in the message (the From: header)
        :param to: primary recipient(s) (displayed in the message)
        :param cc: secondary recipient(s) (displayed in the message)
        :param bcc: secondary recipient(s) (**not** displayed in the message)
        :param charset: character encoding of the message
        :param plain_body: plaintext body
        :param html_body: HTML body

        """
        assert check_argument_types()
        msg = EmailMessage()
        msg['Subject'] = subject or self.message_defaults.get('subject')

        sender = sender or self.message_defaults.get('sender')
        if sender:
            msg['From'] = sender

        to = to or self.message_defaults.get('to')
        if to:
            msg['To'] = to

        cc = cc or self.message_defaults.get('cc')
        if cc:
            msg['Cc'] = cc

        bcc = bcc or self.message_defaults.get('bcc')
        if bcc:
            msg['Bcc'] = bcc

        charset = charset or self.message_defaults.get('charset')
        if plain_body is not None and html_body is not None:
            msg.set_content(plain_body, charset=charset)
            msg.add_alternative(html_body, charset=charset, subtype='html')
        elif plain_body is not None:
            msg.set_content(plain_body, charset=charset)
        elif html_body is not None:
            msg.set_content(html_body, charset=charset, subtype='html')

        return msg

    @classmethod
    def add_attachment(cls, msg: EmailMessage, content: bytes, filename: str,
                       mimetype: str = None):
        """
        Add binary data as an attachment to an :class:`~email.message.EmailMessage`.

        The default value for the ``mimetype`` argument is guessed from the file name.
        If guessing fails, ``application/octet-stream`` is used.

        :param msg: the message
        :param content: the contents of the attachment
        :param filename: the displayed file name in the message
        :param mimetype: the MIME type indicating the type of the file

        """
        assert check_argument_types()
        if not mimetype:
            mimetype, _encoding = guess_type(filename, False)
            if not mimetype:
                mimetype = 'application/octet-stream'

        maintype, subtype = mimetype.split('/', 1)
        if not maintype or not subtype:
            raise ValueError('mimetype must be a string in the "maintype/subtype" format')

        msg.add_attachment(content, maintype=maintype, subtype=subtype, filename=filename)

    @classmethod
    async def add_file_attachment(cls, msg: EmailMessage, path: Union[str, Path],
                                  filename: str = None, mimetype: str = None):
        """
        Read the contents of a file and add them as an attachment to the given message.

        Reads the file contents and then passes the result as ``content`` to
        :meth:`add_attachment` along with the rest of the arguments.

        :param msg: the message
        :param path: path to the file to attach
        :param filename: the displayed file name in the message
        :param mimetype: the MIME type indicating the type of the file

        """
        assert check_argument_types()
        path = Path(path)
        content = await call_in_executor(path.read_bytes)
        cls.add_attachment(msg, content, filename or path.name, mimetype)

    def create_and_deliver(self, **kwargs) -> Awaitable:
        """
        Build a new email message and deliver it.

        This is a shortcut to calling :meth:`create_message` and then passing the result to
        :meth:`deliver`.

        :param kwargs: keyword arguments passed to :meth:`create_message`
        """

        msg = self.create_message(**kwargs)
        return self.deliver(msg)

    @abstractmethod
    async def deliver(self, messages: Union[EmailMessage, Iterable[EmailMessage]]):
        """
        Deliver the given message(s).

        :param messages: the message or iterable of messages to deliver
        """



"""
A simple command line tool that connects to an SMTP server, sends a mail message and then exits.

To run this, you first need to replace the placeholder value for add_component() and
ctx.mailer.create_and_deliver().
"""

import asyncio
import logging

from asphalt.core import ContainerComponent, Context, run_application


class ApplicationComponent(ContainerComponent):
    async def start(self, ctx: Context):
        self.add_component('mailer', backend='smtp', host='smtp.example.org', ssl=True,
                           username='myusername', password='secret')
        await super().start(ctx)

        await ctx.mailer.create_and_deliver(
            subject='Test email', sender='Sender <sender@example.org>',
            to='Recipient <person@example.org>', plain_body='Hello, world!')
        asyncio.get_event_loop().stop()

run_application(ApplicationComponent(), logging=logging.DEBUG)



import sys

import os
import pytest

from asphalt.mailer.api import DeliveryError
from asphalt.mailer.mailers.sendmail import SendmailMailer

pytestmark = pytest.mark.skipif(os.name == 'nt',
                                reason="subprocesses don't work with WindowsSelectorEventLoop")


@pytest.fixture
def outfile(tmpdir):
    return tmpdir.join('outfile')


@pytest.fixture
def script(tmpdir, recipients, outfile):
    p = tmpdir.join('sendmail')
    p.write("""\
#!{interpreter}
import sys

if sys.argv[1:] != ['-i', '-B', '8BITMIME'] + list({recipients!r}):
    print('Wrong arguments; got {{!r}}'.format(sys.argv), file=sys.stderr)
    sys.exit(1)

with open({outfile!r}, 'w') as f:
    f.write(sys.stdin.read())
""".format(interpreter=sys.executable, recipients=recipients, outfile=str(outfile)))
    p.chmod(0o555)
    return str(p)


@pytest.fixture
def fail_script(tmpdir):
    p = tmpdir.join('sendmail')
    p.write("""\
#!{interpreter}
import sys

print('This is a test error', file=sys.stderr)
sys.exit(1)

""".format(interpreter=sys.executable))
    p.chmod(0o555)
    return str(p)


@pytest.fixture
def mailer():
    return SendmailMailer()


@pytest.mark.parametrize('as_list', [True, False])
@pytest.mark.asyncio
async def test_deliver(mailer, script, outfile, as_list, sample_message, recipients):
    mailer.path = script
    await mailer.deliver([sample_message] if as_list else sample_message)
    assert outfile.read() == """\
From: foo@bar.baz
To: Test Recipient <test@domain.country>, test2@domain.country
Cc: Test CC <testcc@domain.country>, testcc2@domain.country
Content-Type: text/plain; charset="utf-8"
Content-Transfer-Encoding: 7bit
MIME-Version: 1.0

Test content
"""


@pytest.mark.asyncio
async def test_deliver_launch_error(mailer, sample_message):
    mailer.path = '/bogus/no/way/this/exists'
    with pytest.raises(DeliveryError) as exc:
        await mailer.deliver(sample_message)

    assert str(exc.value) == ("error sending mail message: "
                              "[Errno 2] No such file or directory: '/bogus/no/way/this/exists'")


@pytest.mark.asyncio
async def test_deliver_error(mailer, sample_message, fail_script):
    mailer.path = fail_script
    with pytest.raises(DeliveryError) as exc:
        await mailer.deliver(sample_message)

    assert str(exc.value) == 'error sending mail message: This is a test error'


def test_repr(mailer):
    assert repr(mailer) == "SendmailMailer('/usr/sbin/sendmail')".format()



import pytest

from asphalt.mailer.mailers.mock import MockMailer


@pytest.fixture
def mailer():
    return MockMailer()


@pytest.mark.asyncio
async def test_deliver(mailer):
    for body in ('message 1', 'message 2'):
        await mailer.create_and_deliver(
            subject='test', sender='foobar@example.org', to='bar@example.org', plain_body=body)

    assert len(mailer.messages) == 2
    assert mailer.messages[0].get_content() == 'message 1\n'
    assert mailer.messages[1].get_content() == 'message 2\n'


def test_repr(mailer):
    assert repr(mailer) == 'MockMailer()'



from pathlib import Path

from setuptools import setup

setup(
    name='asphalt-mailer',
    use_scm_version={
        'version_scheme': 'post-release',
        'local_scheme': 'dirty-tag'
    },
    description='Mailer component for the Asphalt framework',
    long_description=Path(__file__).parent.joinpath('README.rst').read_text('utf-8'),
    author='Alex Grönholm',
    author_email='alex.gronholm@nextday.fi',
    url='https://github.com/asphalt-framework/asphalt-mailer',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Topic :: Communications :: Email',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5'
    ],
    license='Apache License 2.0',
    zip_safe=False,
    packages=[
        'asphalt.mailer',
        'asphalt.mailer.mailers'
    ],
    setup_requires=[
        'setuptools_scm >= 1.7.0'
    ],
    install_requires=[
        'asphalt ~= 2.0'
    ],
    entry_points={
        'asphalt.components': [
            'mailer = asphalt.mailer.component:MailerComponent'
        ],
        'asphalt.mailer.mailers': [
            'mock = asphalt.mailer.mailers.mock:MockMailer',
            'smtp = asphalt.mailer.mailers.smtp:SMTPMailer',
            'sendmail = asphalt.mailer.mailers.sendmail:SendmailMailer'
        ]
    }
)



import re
import ssl
from asyncio import start_server
from asyncio.tasks import Task, gather, wait_for
from functools import partial

import pytest

from asphalt.core.context import Context
from asphalt.mailer.api import DeliveryError
from asphalt.mailer.mailers.smtp import SMTPMailer, SMTPError


class MockSMTPServer:
    def __init__(self, reader, writer, auth_mechanism, fail_on, recipients):
        self.reader = reader
        self.writer = writer
        self.auth_mechanism = auth_mechanism
        self.fail_on = fail_on
        self.recipients = recipients
        self.features = ['8BITMIME', 'PIPELINING']
        if auth_mechanism:
            self.features.append('AUTH %s FOO BAR BAZ' % auth_mechanism)

    async def readline(self):
        read_task = self.reader.readline()
        line = await wait_for(read_task, 1)
        return line.decode().rstrip()

    async def expect(self, pattern: str):
        command = await self.readline()
        match = re.match(pattern, command)
        if not match:
            raise ValueError('Unexpected command: "{}" did not match "{}"'.
                             format(command, pattern))
        groups = match.groups()
        return groups[0] if len(groups) == 1 else groups

    def respond(self, code: int, *lines):
        lines = list(lines)
        for i, line in enumerate(lines):
            separator = '-' if i < (len(lines) - 1) else ' '
            lines[i] = str(code) + separator + line

        output = '\r\n'.join(line for line in lines).encode() + b'\r\n'
        self.writer.write(output)

    @classmethod
    def connected(cls, reader, writer, auth_mechanism, fail_on, recipients):
        server = cls(reader, writer, auth_mechanism, fail_on, recipients)
        return server.protocol()

    async def protocol(self):
        if self.fail_on == 'handshake':
            self.writer.close()
            return
        elif self.fail_on == 'protocol':
            self.writer.write(b'blablabla\r\n')
            await self.writer.drain()
            return
        else:
            self.respond(200, 'fake SMTP server')

        await self.expect('EHLO \w+')
        self.respond(250, 'fake SMTP server', *self.features)

        if self.auth_mechanism == 'LOGIN':
            login, password = await self.expect('AUTH LOGIN (\w+) (\w+)')
            assert login == 'testuser' and password == 'testpass'
        elif self.auth_mechanism == 'PLAIN':
            token = await self.expect('AUTH PLAIN (\w+)')
            assert token == 'AHRlc3R1c2VyAHRlc3RwYXNz'
        else:
            await self.expect('QUIT')
            return

        if self.fail_on == 'auth':
            self.respond(535, 'Authentication failed')
            await self.expect('QUIT')
            return
        else:
            self.respond(235, 'Accepted')

        sender = await self.expect('MAIL FROM: <(.+)>')
        assert sender == 'foo@bar.baz'

        for recipient in self.recipients:
            actual = await self.expect('RCPT TO: <(.+)>')
            assert actual == recipient

        await self.expect('DATA')
        if self.fail_on == 'delivery':
            self.respond(550, 'Rejected')
            return

        self.respond(250, '2.1.0 MAIL ok')
        for recipient in self.recipients:
            self.respond(250, '2.1.5 <{}> ok'.format(recipient))
        self.respond(354, 'Enter mail, end with "." on a line by itself')

        message = []
        while True:
            line = await self.readline()
            if line == '.':
                break
            else:
                message.append(line)

        self.respond(250, '2.6.0 message received')

        await self.expect('QUIT')
        self.respond(221, '2.0.0 goodbye')


@pytest.yield_fixture
def mailer(event_loop, unused_tcp_port, request, recipients):
    auth_mechanism, fail_on = request.param
    callback = partial(MockSMTPServer.connected, auth_mechanism=auth_mechanism,
                       fail_on=fail_on, recipients=recipients)
    task = start_server(callback, '127.0.0.1', unused_tcp_port)
    server = event_loop.run_until_complete(task)
    mailer = SMTPMailer(host='127.0.0.1', port=unused_tcp_port, username='testuser',
                        password='testpass', timeout=1)
    event_loop.run_until_complete(mailer.start(Context()))
    yield mailer
    server.close()
    tasks = Task.all_tasks(event_loop)
    event_loop.run_until_complete(gather(*tasks))


@pytest.mark.asyncio
async def test_resources():
    mailer = SMTPMailer(ssl='contextresource')
    context = Context()
    sslcontext = ssl.create_default_context()
    context.publish_resource(sslcontext, 'contextresource')
    await mailer.start(context)
    assert mailer.ssl is sslcontext


@pytest.mark.parametrize('mailer', [('LOGIN', None), ('PLAIN', None)], indirect=True)
@pytest.mark.asyncio
async def test_deliver(mailer, sample_message):
    await mailer.deliver(sample_message)


@pytest.mark.parametrize('mailer', [(None, None), ('BOGUS', None)], indirect=True)
@pytest.mark.asyncio
async def test_deliver_noauth(mailer, sample_message):
    with pytest.raises(DeliveryError) as exc:
        await mailer.deliver(sample_message)

    assert str(exc.value) == ('error sending mail message: server does not support any '
                              'of our authentication methods')


@pytest.mark.parametrize('mailer', [('LOGIN', 'auth')], indirect=True)
@pytest.mark.asyncio
async def test_deliver_auth_error(mailer, sample_message):
    with pytest.raises(DeliveryError) as exc:
        await mailer.deliver(sample_message)

    assert str(exc.value) == ('error sending mail message: server returned error: '
                              '535 Authentication failed')


@pytest.mark.parametrize('mailer', [('LOGIN', 'handshake')], indirect=True)
@pytest.mark.asyncio
async def test_deliver_handshake_error(mailer, sample_message):
    with pytest.raises(DeliveryError) as exc:
        await mailer.deliver(sample_message)

    assert str(exc.value) == 'error sending mail message: server closed connection'


@pytest.mark.parametrize('mailer', [('LOGIN', 'protocol')], indirect=True)
@pytest.mark.asyncio
async def test_deliver_protocol_error(mailer, sample_message):
    with pytest.raises(DeliveryError) as exc:
        await mailer.deliver(sample_message)

    assert str(exc.value) == ('error sending mail message: SMTP protocol error: '
                              'received unexpected response: blablabla')


@pytest.mark.parametrize('mailer', [('LOGIN', 'delivery')], indirect=True)
@pytest.mark.asyncio
async def test_deliver_delivery_error(mailer, sample_message):
    with pytest.raises(SMTPError) as exc:
        await mailer.deliver(sample_message)

    assert exc.value.code == 550
    assert exc.value.message is sample_message
    # assert str(exc.value) == ('error sending mail message: SMTP protocol error: '
    #                           'received unexpected response: blablabla')


@pytest.mark.parametrize('mailer', [('LOGIN', None)], indirect=True)
def test_repr(mailer):
    assert repr(mailer).startswith("SMTPMailer(host='127.0.0.1'")



"""
A simple command line tool that connects to an SMTP server, sends a mail message with an attachment
and then exits.

To run this, you first need to replace the placeholder value for add_component() and
ctx.mailer.create_and_deliver().

The program requires one command line argument, the path to the attachment file.
"""

import asyncio
import logging
import os.path
import sys

from asphalt.core import ContainerComponent, Context, run_application


class ApplicationComponent(ContainerComponent):
    def __init__(self, attachment):
        super().__init__()
        self.attachment = attachment

    async def start(self, ctx: Context):
        self.add_component('mailer', backend='smtp', host='smtp.example.org', ssl=True,
                           username='myusername', password='secret')
        await super().start(ctx)

        message = ctx.mailer.create_message(
            subject='Test email with attachment', sender='Sender <sender@example.org>',
            to='Recipient <person@example.org>', plain_body='Take a look at this file!')
        await ctx.mailer.add_file_attachment(message, self.attachment)
        await ctx.mailer.deliver(message)
        asyncio.get_event_loop().stop()


if len(sys.argv) < 2:
    print('Specify the path to the attachment as the argument to this script!', file=sys.stderr)
    sys.exit(1)

if not os.path.isfile(sys.argv[1]):
    print('The attachment ({}) does not exist or is not a regular file', file=sys.stderr)
    sys.exit(2)

run_application(ApplicationComponent(sys.argv[1]), logging=logging.DEBUG)
